/*
  ==============================================================================

   This file is part of the JUCE library.
   Copyright (c) 2020 - Raw Material Software Limited

   JUCE is an open source library subject to commercial or open-source
   licensing.

   By using JUCE, you agree to the terms of both the JUCE 6 End-User License
   Agreement and JUCE Privacy Policy (both effective as of the 16th June 2020).

   End User License Agreement: www.juce.com/juce-6-licence
   Privacy Policy: www.juce.com/juce-privacy-policy

   Or: You may also use this code under the terms of the GPL v3 (see
   www.gnu.org/licenses).

   JUCE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY, AND ALL WARRANTIES, WHETHER
   EXPRESSED OR IMPLIED, INCLUDING MERCHANTABILITY AND FITNESS FOR PURPOSE, ARE
   DISCLAIMED.

  ==============================================================================
*/

COL(black)
COL(white)
COL(blue)
COL(grey)
COL(green)
COL(red)
COL(yellow)
COL(aliceblue)
COL(antiquewhite)
COL(aqua)
COL(aquamarine)
COL(azure)
COL(beige)
COL(bisque)
COL(blanchedalmond)
COL(blueviolet)
COL(brown)
COL(burlywood)
COL(cadetblue)
COL(chartreuse)
COL(chocolate)
COL(coral)
COL(cornflowerblue)
COL(cornsilk)
COL(crimson)
COL(cyan)
COL(darkblue)
COL(darkcyan)
COL(darkgoldenrod)
COL(darkgrey)
COL(darkgreen)
COL(darkkhaki)
COL(darkmagenta)
COL(darkolivegreen)
COL(darkorange)
COL(darkorchid)
COL(darkred)
COL(darksalmon)
COL(darkseagreen)
COL(darkslateblue)
COL(darkslategrey)
COL(darkturquoise)
COL(darkviolet)
COL(deeppink)
COL(deepskyblue)
COL(dimgrey)
COL(dodgerblue)
COL(firebrick)
COL(floralwhite)
COL(forestgreen)
COL(fuchsia)
COL(gainsboro)
COL(gold)
COL(goldenrod)
COL(greenyellow)
COL(honeydew)
COL(hotpink)
COL(indianred)
COL(indigo)
COL(ivory)
COL(khaki)
COL(lavender)
COL(lavenderblush)
COL(lemonchiffon)
COL(lightblue)
COL(lightcoral)
COL(lightcyan)
COL(lightgoldenrodyellow)
COL(lightgreen)
COL(lightgrey)
COL(lightpink)
COL(lightsalmon)
COL(lightseagreen)
COL(lightskyblue)
COL(lightslategrey)
COL(lightsteelblue)
COL(lightyellow)
COL(lime)
COL(limegreen)
COL(linen)
COL(magenta)
COL(maroon)
COL(mediumaquamarine)
COL(mediumblue)
COL(mediumorchid)
COL(mediumpurple)
COL(mediumseagreen)
COL(mediumslateblue)
COL(mediumspringgreen)
COL(mediumturquoise)
COL(mediumvioletred)
COL(midnightblue)
COL(mintcream)
COL(mistyrose)
COL(navajowhite)
COL(navy)
COL(oldlace)
COL(olive)
COL(olivedrab)
COL(orange)
COL(orangered)
COL(orchid)
COL(palegoldenrod)
COL(palegreen)
COL(paleturquoise)
COL(palevioletred)
COL(papayawhip)
COL(peachpuff)
COL(peru)
COL(pink)
COL(plum)
COL(powderblue)
COL(purple)
COL(rosybrown)
COL(royalblue)
COL(saddlebrown)
COL(salmon)
COL(sandybrown)
COL(seagreen)
COL(seashell)
COL(sienna)
COL(silver)
COL(skyblue)
COL(slateblue)
COL(slategrey)
COL(snow)
COL(springgreen)
COL(steelblue)
COL(tan)
COL(teal)
COL(thistle)
COL(tomato)
COL(turquoise)
COL(violet)
COL(wheat)
COL(whitesmoke)
COL(yellowgreen)
